"use strict";
// Manages ~/.fauna-shell
Object.defineProperty(exports, "__esModule", { value: true });
exports.fileExists = exports.getProjectConfigPath = exports.getRootConfigPath = exports.ShellConfig = exports.Config = exports.InvalidConfigError = exports.PROJECT_FILE_NAME = exports.Stack = exports.Endpoint = exports.ProjectConfig = exports.RootConfig = void 0;
const tslib_1 = require("tslib");
const fs_1 = tslib_1.__importDefault(require("fs"));
const os_1 = tslib_1.__importDefault(require("os"));
const path_1 = tslib_1.__importDefault(require("path"));
const ini = require("ini");
const project_config_1 = require("./project-config");
Object.defineProperty(exports, "ProjectConfig", { enumerable: true, get: function () { return project_config_1.ProjectConfig; } });
Object.defineProperty(exports, "Stack", { enumerable: true, get: function () { return project_config_1.Stack; } });
const root_config_1 = require("./root-config");
Object.defineProperty(exports, "RootConfig", { enumerable: true, get: function () { return root_config_1.RootConfig; } });
Object.defineProperty(exports, "Endpoint", { enumerable: true, get: function () { return root_config_1.Endpoint; } });
const secret_1 = require("../secret");
exports.PROJECT_FILE_NAME = ".fauna-project";
class InvalidConfigError extends Error {
}
exports.InvalidConfigError = InvalidConfigError;
// Wraps an `ini` file with helpers to get typed values out
class Config {
    constructor(keyName, config) {
        this.keyName = keyName;
        this.config = config;
    }
    strOpt(key) {
        const v = this.config[key];
        if (v === undefined || typeof v === "string") {
            return v;
        }
        else {
            throw new InvalidConfigError(`Expected string for ${this.keyName} ${key}, got ${typeof v}`);
        }
    }
    keys() {
        return Object.keys(this.config);
    }
    numberOpt(key) {
        const v = this.config[key];
        if (v === undefined || typeof v === "number") {
            return v;
        }
        else if (typeof v === "string") {
            try {
                return parseInt(v, 10);
            }
            catch (_) {
                throw new InvalidConfigError(`Invalid number for ${this.keyName} ${key}`);
            }
        }
        else {
            throw new InvalidConfigError(`Expected number for ${this.keyName} ${key}, got ${typeof v}`);
        }
    }
    object(key) {
        const v = this.config[key];
        if (v === undefined) {
            throw new InvalidConfigError(`Missing value for required ${this.keyName} ${key}`);
        }
        else if (v !== null && typeof v === "object") {
            return new Config(this.keyName, v);
        }
        else {
            throw new InvalidConfigError(`Expected object for ${this.keyName} ${key}, got ${typeof v}`);
        }
    }
    objectExists(key) {
        const v = this.config[key];
        return v !== undefined && typeof v === "object";
    }
    str(key) {
        return this.require(key, this.strOpt(key));
    }
    number(key) {
        return this.require(key, this.numberOpt(key));
    }
    require(key, value) {
        if (value === undefined) {
            throw new InvalidConfigError(`Missing value for required ${this.keyName} ${key}`);
        }
        else {
            return value;
        }
    }
    // Returns a list of all keys that match `pred`.
    allObjectsWhere(pred) {
        return Object.keys(this.config).flatMap((k) => pred(k) ? [[k, this.object(k)]] : []);
    }
    // Returns a list of all child objects of the given `key`.
    objectsIn(key) {
        const obj = this.object(key);
        return obj.allObjectsWhere((_) => true);
    }
}
exports.Config = Config;
class ShellConfig {
    static read(flags, log) {
        const rootConfig = ini.parse(readFileOpt((0, exports.getRootConfigPath)()));
        const projectConfigPath = (0, exports.getProjectConfigPath)();
        const projectConfig = projectConfigPath
            ? ini.parse(readFile(projectConfigPath))
            : undefined;
        const shellConfig = new ShellConfig({
            flags,
            rootConfig,
            projectPath: projectConfigPath !== undefined
                ? path_1.default.dirname(projectConfigPath)
                : undefined,
            projectConfig,
        });
        if (log !== undefined) {
            shellConfig.configErrors().forEach((err) => log.warn(err));
        }
        return shellConfig;
    }
    static readWithOverrides(opts) {
        var _a, _b, _c;
        const rootConfig = (_a = opts === null || opts === void 0 ? void 0 : opts.rootConfig) !== null && _a !== void 0 ? _a : ini.parse(readFileOpt((0, exports.getRootConfigPath)()));
        const projectConfigPath = (_b = opts === null || opts === void 0 ? void 0 : opts.projectPath) !== null && _b !== void 0 ? _b : (0, exports.getProjectConfigPath)();
        const projectConfig = (_c = opts === null || opts === void 0 ? void 0 : opts.projectConfig) !== null && _c !== void 0 ? _c : (projectConfigPath ? ini.parse(readFile(projectConfigPath)) : undefined);
        return new ShellConfig({
            rootConfig,
            projectPath: projectConfigPath !== undefined
                ? path_1.default.dirname(projectConfigPath)
                : undefined,
            projectConfig,
        });
    }
    constructor(opts) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j;
        this.validate = () => {
            if (this.endpoint === undefined) {
                // No `~/.fauna-shell` was found, and no `--secret` was passed.
                throw new Error(`No endpoint or secret set. Set an endpoint in ~/.fauna-shell, ${exports.PROJECT_FILE_NAME}, or pass --endpoint`);
            }
        };
        this.lookupEndpoint = (opts) => {
            var _a, _b;
            this.validate();
            let database = (_b = (_a = this.stack) === null || _a === void 0 ? void 0 : _a.database.split("/")) !== null && _b !== void 0 ? _b : [];
            if (opts.scope !== undefined) {
                database.push(...opts.scope.split("/"));
            }
            return this.endpoint.makeScopedEndpoint(database);
        };
        this.flags = new Config("flag", (_a = opts.flags) !== null && _a !== void 0 ? _a : {});
        this.rootConfig = new root_config_1.RootConfig(new Config("config key", (_b = opts.rootConfig) !== null && _b !== void 0 ? _b : {}));
        this.projectPath = opts.projectPath;
        this.projectConfig = opts.projectConfig
            ? project_config_1.ProjectConfig.fromConfig(new Config("config key", opts.projectConfig))
            : undefined;
        (_c = this.projectConfig) === null || _c === void 0 ? void 0 : _c.validate(this.rootConfig);
        const urlFlag = root_config_1.Endpoint.getURLFromFlags(this.flags);
        if (urlFlag !== undefined) {
            try {
                // eslint-disable-next-line no-new
                new URL(urlFlag);
            }
            catch (e) {
                throw new Error(`Invalid database URL: ${urlFlag}`);
            }
        }
        if (this.projectConfig === undefined) {
            const stackName = this.flags.strOpt("stack");
            if (stackName !== undefined) {
                throw new Error(`No ${exports.PROJECT_FILE_NAME} was found, so stack '${stackName}' cannot be used`);
            }
        }
        else {
            const stackName = (_d = this.flags.strOpt("stack")) !== null && _d !== void 0 ? _d : this.projectConfig.defaultStack;
            if (stackName !== undefined) {
                this.stack = this.projectConfig.stacks[stackName];
                if (this.stack === undefined) {
                    throw new Error(`No such stack '${stackName}'`);
                }
            }
        }
        // An endpoint must be chosen as well. The endpoint may come from (in order)
        // the `--endpoint` flag, the `endpoint` key in the project config, or the
        // `default` key in the root config.
        const endpointName = (_g = (_e = this.flags.strOpt("endpoint")) !== null && _e !== void 0 ? _e : (_f = this.stack) === null || _f === void 0 ? void 0 : _f.endpoint) !== null && _g !== void 0 ? _g : this.rootConfig.defaultEndpoint;
        const secretFlag = this.flags.strOpt("secret");
        if (endpointName === undefined) {
            // No `~/.fauna-shell` was found, so `--secret` is required to make an endpoint. If `--secret` wasn't passed, `validate` should fail.
            if (secretFlag !== undefined) {
                this.endpoint = new root_config_1.Endpoint({
                    secret: secret_1.Secret.parse(secretFlag),
                    url: urlFlag,
                    graphqlHost: this.flags.strOpt("graphqlHost"),
                    graphqlPort: this.flags.numberOpt("graphqlPort"),
                });
            }
        }
        else {
            this.endpoint = this.rootConfig.endpoints[endpointName];
            if (this.endpoint === undefined) {
                throw new Error(`No such endpoint '${endpointName}'`);
            }
            // override endpoint with values from flags.
            if (secretFlag !== undefined) {
                this.endpoint.secret = secret_1.Secret.parse(secretFlag);
            }
            this.endpoint.url = urlFlag !== null && urlFlag !== void 0 ? urlFlag : this.endpoint.url;
            this.endpoint.graphqlHost =
                (_h = this.flags.strOpt("graphqlHost")) !== null && _h !== void 0 ? _h : this.endpoint.graphqlHost;
            this.endpoint.graphqlPort =
                (_j = this.flags.numberOpt("graphqlHost")) !== null && _j !== void 0 ? _j : this.endpoint.graphqlPort;
        }
    }
    configErrors() {
        if (this.rootConfig.invalidEndpoints.length > 0) {
            return [
                `The following endpoint definitions in ${(0, exports.getRootConfigPath)()} are invalid:\n ${this.rootConfig.invalidEndpoints.join("\n")}\n Resolve them by ensuring they have a secret defined or remove them if they are not needed.`,
            ];
        }
        else {
            return [];
        }
    }
    /**
     * Saves the project config, if present.
     */
    saveProjectConfig() {
        var _a;
        (_a = this.projectConfig) === null || _a === void 0 ? void 0 : _a.save(this.projectConfigFile());
    }
    /**
     * Saves the root config.
     */
    saveRootConfig() {
        if (this.rootConfig.invalidEndpoints.length > 0) {
            throw new Error(`The following endpoint definitions in ${(0, exports.getRootConfigPath)()} are invalid:\n ${this.rootConfig.invalidEndpoints.join("\n")}\n Resolve them by ensuring they have a secret defined or remove them if they are not needed.`);
        }
        this.rootConfig.save((0, exports.getRootConfigPath)());
    }
    projectConfigFile() {
        return this.projectPath === undefined
            ? undefined
            : path_1.default.join(this.projectPath, exports.PROJECT_FILE_NAME);
    }
}
exports.ShellConfig = ShellConfig;
const readFileOpt = (fileName) => {
    if (fs_1.default.existsSync(fileName)) {
        return fs_1.default.readFileSync(fileName, "utf8");
    }
    else {
        return "";
    }
};
const readFile = (fileName) => {
    return fs_1.default.readFileSync(fileName, "utf8");
};
const getRootConfigPath = () => {
    return path_1.default.join(os_1.default.homedir(), ".fauna-shell");
};
exports.getRootConfigPath = getRootConfigPath;
const getProjectConfigPath = (start) => {
    let current = start;
    if (current === undefined) {
        try {
            current = process.cwd();
        }
        catch (err) {
            // If the cwd is not accessible, just give up. If this happens, one of our
            // dependencies actually explodes elsewhere, but we might as well handle
            // errors where possible.
            return undefined;
        }
    }
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const projPath = path_1.default.join(current, exports.PROJECT_FILE_NAME);
        if ((0, exports.fileExists)(projPath)) {
            return projPath;
        }
        const currPath = path_1.default.parse(current);
        // break if at root
        if (currPath.base === "") {
            break;
        }
        current = currPath.dir;
    }
    // if we got here, it means that there was no `.fauna-project` file, so we
    // give up.
    return undefined;
};
exports.getProjectConfigPath = getProjectConfigPath;
const fileExists = (filePath) => {
    const stat = fs_1.default.statSync(filePath, {
        // returns undefined instead of throwing if the file doesn't exist
        throwIfNoEntry: false,
    });
    return stat !== undefined && stat.isFile();
};
exports.fileExists = fileExists;
