"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StackFactory = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@oclif/core");
const prompts_1 = require("@inquirer/prompts");
const search_select_1 = require("./search-select");
const fauna_client_1 = tslib_1.__importDefault(require("./fauna-client"));
class StackFactory {
    constructor(cmd, config) {
        this.promptDatabasePath = async (endpoint) => {
            const { url, secret } = endpoint;
            const client = new fauna_client_1.default({
                endpoint: url,
                secret: secret.buildSecret(),
            });
            const res = await client.query("0");
            if (res.status !== 200) {
                this.cmd.error(`${res.body.error.code}`);
            }
            const databasePaths = await this.getDatabasePaths(client);
            if (databasePaths === undefined) {
                const res = await (0, prompts_1.input)({
                    message: "Enter a database path",
                    validate: async (path) => {
                        if (await this.validateDatabasePath(client, path)) {
                            return true;
                        }
                        else {
                            return `No such database ${path}`;
                        }
                    },
                });
                await client.close();
                return res;
            }
            else if (databasePaths.length === 0) {
                this.cmd.error("No databases found in the given endpoint");
            }
            else {
                await client.close();
                return (0, search_select_1.searchSelect)({
                    message: "Select a database",
                    choices: databasePaths.map((database) => ({
                        value: database,
                    })),
                });
            }
        };
        this.validateDatabasePath = async (client, path) => {
            const res = await client.query(`0`, {
                secret: `${client.secret}:${path}:admin`,
            });
            return res.status === 200;
        };
        this.getDatabasePaths = async (client) => {
            // Limits: choose a limit of 100 databases at each depth, and a depth of 3.
            // We will also add a limit if any databases are skiped.
            const databaseLimit = 100;
            const depthLimit = 3;
            const allDatabases = [];
            const paths = [[]];
            let overflowedLimit = false;
            let overflowedDepth = false;
            core_1.ux.action.start("Fetching databases");
            // eslint-disable-next-line no-constant-condition
            while (!overflowedLimit && !overflowedDepth) {
                const path = paths.pop();
                if (path === undefined) {
                    break;
                }
                if (path.length > depthLimit) {
                    overflowedDepth = true;
                    continue;
                }
                const databases = await client.query(`Database.all().take(${databaseLimit + 1}).toArray().map(.name)`, {
                    secret: `${client.secret}:${path.join("/")}:admin`,
                });
                if (databases.status !== 200) {
                    this.cmd.error(`Error: ${databases.body.error.code}`);
                }
                const dbs = databases.body.data;
                if (dbs.length > databaseLimit) {
                    overflowedLimit = true;
                }
                const nestedPaths = dbs.map((database) => [...path, database]);
                paths.push(...nestedPaths);
                allDatabases.push(...nestedPaths.map((path) => path.join("/")));
            }
            core_1.ux.action.stop();
            if (overflowedDepth || overflowedLimit) {
                this.cmd.log("Note: there are too many databases to display in this endpoint");
                return undefined;
            }
            else {
                allDatabases.sort();
                return allDatabases;
            }
        };
        this.cmd = cmd;
        this.config = config;
    }
    async askEndpoint() {
        const useDefault = this.config.rootConfig.defaultEndpoint !== undefined &&
            (await (0, prompts_1.confirm)({
                message: `Use the default endpoint [${this.config.rootConfig.defaultEndpoint}]`,
            }));
        if (useDefault) {
            return this.config.rootConfig.defaultEndpoint;
        }
        return (0, search_select_1.searchSelect)({
            message: "Select an endpoint",
            choices: Object.keys(this.config.rootConfig.endpoints).map((endpoint) => ({
                value: endpoint,
            })),
        });
    }
    /**
     *
     * @param cmd - used to send info/errors to the user.
     * This method will be executed from the project init as well as stack add method
     * @param config
     */
    async addStack(params) {
        var _a, _b, _c, _d;
        if ((params === null || params === void 0 ? void 0 : params.endpoint) === undefined &&
            Object.keys(this.config.rootConfig.endpoints).length === 0) {
            this.cmd.error("No endpoints found. Use `fauna cloud-login` to configure an endpoint.");
        }
        const validateStackName = (input) => {
            var _a, _b;
            if (input.length === 0) {
                return "Stack name cannot be empty";
            }
            else if (Object.keys((_b = (_a = this.config.projectConfig) === null || _a === void 0 ? void 0 : _a.stacks) !== null && _b !== void 0 ? _b : {}).includes(input)) {
                return `Stack ${input} already exists`;
            }
            else {
                return true;
            }
        };
        const name = (_a = params === null || params === void 0 ? void 0 : params.name) !== null && _a !== void 0 ? _a : (await (0, prompts_1.input)({
            message: "Stack name",
            validate: validateStackName,
        }));
        const res = validateStackName(name);
        if (res !== true) {
            this.cmd.error(res);
        }
        const endpointName = (_b = params === null || params === void 0 ? void 0 : params.endpoint) !== null && _b !== void 0 ? _b : (await this.askEndpoint());
        if (!Object.keys(this.config.rootConfig.endpoints).includes(endpointName)) {
            this.cmd.error(`No such endpoint '${endpointName}'`);
        }
        let databaseName = (_c = params === null || params === void 0 ? void 0 : params.database) !== null && _c !== void 0 ? _c : (await this.promptDatabasePath(this.config.rootConfig.endpoints[endpointName]));
        const setDefault = (_d = params === null || params === void 0 ? void 0 : params.default) !== null && _d !== void 0 ? _d : ((params === null || params === void 0 ? void 0 : params.nonInteractive)
            ? false
            : await (0, prompts_1.confirm)({
                message: "Make this stack default",
            }));
        if (setDefault) {
            this.config.projectConfig.defaultStack = name;
        }
        this.config.projectConfig.stacks[name] = {
            endpoint: endpointName,
            database: databaseName,
        };
        this.config.saveProjectConfig();
        console.log(`Saved stack ${name} to ${this.config.projectConfigFile()}`);
    }
}
exports.StackFactory = StackFactory;
